import spacy
from spacy.tokens import Span
from spacy.strings import StringStore

from .core import QuickUMLS
from . import constants

class SpacyQuickUMLS(object):
    name = 'QuickUMLS matcher'
    
    def __init__(self, nlp, quickumls_fp, best_match=True, ignore_syntax=False, **kwargs):
        """Instantiate SpacyQuickUMLS object

            This creates a QuickUMLS spaCy component which can be used in modular pipelines.  
            This module adds entity Spans to the document where the entity label is the UMLS CUI and the Span's "underscore" object is extended to contains "similarity" and "semtypes" for matched concepts.
            Note that this implementation follows and enforces a known spacy convention that entity Spans cannot overlap on a single token.

        Args:
            nlp: Existing spaCy pipeline.  This is needed to update the vocabulary with UMLS CUI values
            quickumls_fp (str): Path to QuickUMLS data
            best_match (bool, optional): Whether to return only the top match or all overlapping candidates. Defaults to True.
            ignore_syntax (bool, optional): Whether to use the heuristcs introduced in the paper (Soldaini and Goharian, 2016). TODO: clarify,. Defaults to False
            **kwargs: QuickUMLS keyword arguments (see QuickUMLS in core.py)
        """
        
        self.quickumls = QuickUMLS(quickumls_fp, 
            # By default, the QuickUMLS objects creates its own internal spacy pipeline but this is not needed
            # when we're using it as a component in a pipeline
            spacy_component = True,
            **kwargs)
        
        # save this off so that we can get vocab values of labels later
        self.nlp = nlp
        
        # keep these for matching
        self.best_match = best_match
        self.ignore_syntax = ignore_syntax

        # let's extend this with some proprties that we want
        Span.set_extension('similarity', default = -1.0)
        Span.set_extension('semtypes', default = -1.0)
        
    def __call__(self, doc):
        # pass in the document which has been parsed to this point in the pipeline for ngrams and matches
        matches = self.quickumls._match(doc, best_match=self.best_match, ignore_syntax=self.ignore_syntax)
        
        # NOTE: Spacy spans do not allow overlapping tokens, so we prevent the overlap here
        # For more information, see: https://github.com/explosion/spaCy/issues/3608
        tokens_in_ents_set = set()
        
        # let's track any other entities which may have been attached via upstream components
        for ent in doc.ents:
            for token_index in range(ent.start, ent.end):
                tokens_in_ents_set.add(token_index)
        
        # Convert QuickUMLS match objects into Spans
        for match in matches:
            # each match may match multiple ngrams
            for ngram_match_dict in match:
                start_char_idx = int(ngram_match_dict['start'])
                end_char_idx = int(ngram_match_dict['end'])
                
                cui = ngram_match_dict['cui']
                # add the string to the spacy vocab
                self.nlp.vocab.strings.add(cui)
                # pull out the value
                cui_label_value = self.nlp.vocab.strings[cui]
                
                # char_span() creates a Span from these character indices
                # UMLS CUI should work well as the label here
                span = doc.char_span(start_char_idx, end_char_idx, label = cui_label_value)
                
                # before we add this, let's make sure that this entity does not overlap any tokens added thus far
                candidate_token_indexes = set(range(span.start, span.end))
                
                # check the intersection and skip this if there is any overlap
                if len(tokens_in_ents_set.intersection(candidate_token_indexes)) > 0:
                    continue
                    
                # track this to make sure we do not introduce overlap later
                tokens_in_ents_set.update(candidate_token_indexes)
                
                # add some custom metadata to the spans
                span._.similarity = ngram_match_dict['similarity']
                span._.semtypes = ngram_match_dict['semtypes']
                doc.ents = list(doc.ents) + [span]
                
        return doc


# inspired from:
# https://python-packaging-user-guide.readthedocs.org/en/latest/single_source_version/
# https://github.com/pypa/warehouse/blob/master/warehouse/__about__.py
# https://github.com/explosion/spaCy/blob/master/spacy/about.py

__title__ = 'medspacy_quickumls'
__version__ = '2.2'
__author__ = 'Luca Soldaini, modified by Kelly Peterson'
__email__ = 'luca@ir.cs.georgetown.edu,kelly.peterson@hsc.utah.edu'
__license__ = 'MIT'
__uri__ = "https://github.com/medspacy/quickumls"
__copyright__ = '2014-2020, Georgetown University Information Retrieval Lab, updated 2020, University of Utah'



from __future__ import unicode_literals, division, print_function

# built in modules
import argparse
import codecs
import os
from six.moves import input
import shutil
import sys
import time
try:
    from unidecode import unidecode
except ImportError:
    pass


# third party-dependencies
import spacy


# project modules
from .toolbox import countlines, CuiSemTypesDB, SimstringDBWriter, mkdir
from .constants import HEADERS_MRCONSO, HEADERS_MRSTY, LANGUAGES, SPACY_LANGUAGE_MAP


def get_semantic_types(path, headers):
    sem_types = {}
    with codecs.open(path, encoding='utf-8') as f:
        for i, ln in enumerate(f):
            content = dict(zip(headers, ln.strip().split('|')))

            sem_types.setdefault(content['cui'], []).append(content['sty'])

    return sem_types


def get_mrconso_iterator(path, headers, lang='ENG'):
    with codecs.open(path, encoding='utf-8') as f:
        for i, ln in enumerate(f):
            content = dict(zip(headers, ln.strip().split('|')))

            if content['lat'] != lang:
                continue

            yield content


def extract_from_mrconso(
        mrconso_path, mrsty_path, opts,
        mrconso_header=HEADERS_MRCONSO, mrsty_header=HEADERS_MRSTY):

    start = time.time()
    print('loading semantic types...', end=' ')
    sys.stdout.flush()
    sem_types = get_semantic_types(mrsty_path, mrsty_header)
    print('done in {:.2f} s'.format(time.time() - start))

    start = time.time()

    mrconso_iterator = get_mrconso_iterator(
        mrconso_path, mrconso_header, opts.language
    )

    total = countlines(mrconso_path)

    processed = set()
    i = 0

    for content in mrconso_iterator:
        i += 1

        if i % 100000 == 0:
            delta = time.time() - start
            status = (
                '{:,} in {:.2f} s ({:.2%}, {:.1e} s / term)'
                ''.format(i, delta, i / total, delta / i if i > 0 else 0)
            )
            print(status)

        concept_text = content['str'].strip()
        cui = content['cui']
        preferred = 1 if content['ispref'] == 'Y' else 0

        if opts.lowercase:
            concept_text = concept_text.lower()

        if opts.normalize_unicode:
            concept_text = unidecode(concept_text)

        if (cui, concept_text) in processed:
            continue
        else:
            processed.add((cui, concept_text))

        yield (concept_text, cui, sem_types[cui], preferred)

    delta = time.time() - start
    status = (
        '\nCOMPLETED: {:,} in {:.2f} s ({:.1e} s / term)'
        ''.format(i, delta, i / total, delta / i if i > 0 else 0)
    )
    print(status)


def parse_and_encode_ngrams(extracted_it, simstring_dir, cuisty_dir, database_backend):
    # Create destination directories for the two databases
    mkdir(simstring_dir)
    mkdir(cuisty_dir)

    ss_db = SimstringDBWriter(simstring_dir)
    cuisty_db = CuiSemTypesDB(cuisty_dir, database_backend=database_backend)

    simstring_terms = set()

    for i, (term, cui, stys, preferred) in enumerate(extracted_it, start=1):
        if term not in simstring_terms:
            ss_db.insert(term)
            simstring_terms.add(term)

        cuisty_db.insert(term, cui, stys, preferred)


def install_spacy(lang):
    """Tries to create a spacy object; if it fails, downloads the dataset"""

    print(f'Determining if SpaCy for language "{lang}" is installed...')

    if lang in SPACY_LANGUAGE_MAP:
        try:
            spacy.load(SPACY_LANGUAGE_MAP[lang])
            print(f'SpaCy is installed and avaliable for {lang}!')
        except OSError:
            print(f'SpaCy is not available! Attempting to download and install...')
            spacy.cli.download(SPACY_LANGUAGE_MAP[lang])


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        'umls_installation_path',
        help=('Location of UMLS installation files (`MRCONSO.RRF` and '
              '`MRSTY.RRF` files)')
    )
    ap.add_argument(
        'destination_path',
        help='Location where the necessary QuickUMLS files are installed'
    )
    ap.add_argument(
        '-L', '--lowercase', action='store_true',
        help='Consider only lowercase version of tokens'
    )
    ap.add_argument(
        '-U', '--normalize-unicode', action='store_true',
        help='Normalize unicode strings to their closest ASCII representation'
    )
    ap.add_argument(
        '-d', '--database-backend', choices=('leveldb', 'unqlite'), default='unqlite',
        help='KV database to use to store CUIs and semantic types'
    )
    ap.add_argument(
        '-E', '--language', default='ENG', choices=LANGUAGES,
        help='Extract concepts of the specified language'
    )
    opts = ap.parse_args()
    return opts


def main():
    opts = parse_args()

    install_spacy(opts.language)

    if not os.path.exists(opts.destination_path):
        msg = ('Directory "{}" does not exists; should I create it? [y/N] '
               ''.format(opts.destination_path))
        create = input(msg).lower().strip() == 'y'

        if create:
            os.makedirs(opts.destination_path)
        else:
            print('Aborting.')
            exit(1)

    if len(os.listdir(opts.destination_path)) > 0:
        msg = ('Directory "{}" is not empty; should I empty it? [y/N] '
               ''.format(opts.destination_path))
        empty = input(msg).lower().strip() == 'y'
        if empty:
            shutil.rmtree(opts.destination_path)
            os.mkdir(opts.destination_path)
        else:
            print('Aborting.')
            exit(1)

    if opts.normalize_unicode:
        try:
            unidecode
        except NameError:
            err = ('`unidecode` is needed for unicode normalization'
                   'please install it via the `[sudo] pip install '
                   'unidecode` command.')
            print(err, file=sys.stderr)
            exit(1)

        flag_fp = os.path.join(opts.destination_path, 'normalize-unicode.flag')
        open(flag_fp, 'w').close()

    if opts.lowercase:
        flag_fp = os.path.join(opts.destination_path, 'lowercase.flag')
        open(flag_fp, 'w').close()

    flag_fp = os.path.join(opts.destination_path, 'language.flag')
    with open(flag_fp, 'w') as f:
        f.write(opts.language)

    flag_fp = os.path.join(opts.destination_path, 'database_backend.flag')
    with open(flag_fp, 'w') as f:
        f.write(opts.database_backend)

    mrconso_path = os.path.join(opts.umls_installation_path, 'MRCONSO.RRF')
    mrsty_path = os.path.join(opts.umls_installation_path, 'MRSTY.RRF')

    mrconso_iterator = extract_from_mrconso(mrconso_path, mrsty_path, opts)

    simstring_dir = os.path.join(opts.destination_path, 'umls-simstring.db')
    cuisty_dir = os.path.join(opts.destination_path, 'cui-semtypes.db')

    parse_and_encode_ngrams(mrconso_iterator, simstring_dir, cuisty_dir,
                            database_backend=opts.database_backend)


if __name__ == '__main__':
    main()



from __future__ import unicode_literals, division, print_function

# build-in modules
import re
import os
from functools import wraps
import six
import unicodedata
from string import punctuation
from itertools import takewhile, repeat
from six.moves import xrange

# installed modules
import numpy
import leveldb
try:
    import unqlite
    UNQLITE_AVAILABLE = True
except ImportError:
    UNQLITE_AVAILABLE = False

# project imports
from quickumls_simstring import simstring


# Python version specific imports
if six.PY2:
    import cPickle as pickle
else:
    import pickle


def pickle_loading(data):
    pass


def pickle_dumping(data):
    pass


def mkdir(path):
    try:
        os.makedirs(path)
        return True
    except OSError:
        return False


def count_ngrams(s, n):
    return len(s) + n - 1


def safe_unicode(s):
    if six.PY2:
        # in python 3, there no ambiguity on whether
        # a string is encoded in bytes format or not
        try:
            s = u'%s' % s
        except UnicodeDecodeError:
            s = u'%s' % s.decode('utf-8')

    return u'{}'.format(unicodedata.normalize('NFKD', s))


def prepare_string_for_db_input(s):
    if six.PY2:
        return s.encode('utf-8')
    else:
        return s


def make_ngrams(s, n):
    # s = u'{t}{s}{t}'.format(s=safe_unicode(s), t=('$' * (n - 1)))
    n = len(s) if len(s) < n else n
    return (s[i:i + n] for i in xrange(len(s) - n + 1))


def get_similarity(x, y, n, similarity_name):
    if len(x) == 0 or len(y) == 0:
        # we define similarity between two strings
        # to be 0 if any of the two is empty.
        return 0.

    X, Y = set(make_ngrams(x, n)), set(make_ngrams(y, n))
    intersec = len(X.intersection(Y))

    if similarity_name == 'dice':
        return 2 * intersec / (len(X) + len(Y))
    elif similarity_name == 'jaccard':
        return intersec / (len(X) + len(Y) - intersec)
    elif similarity_name == 'cosine':
        return intersec / numpy.sqrt(len(X) * len(Y))
    elif similarity_name == 'overlap':
        return intersec
    else:
        msg = 'Similarity {} not recognized'.format(similarity_name)
        raise TypeError(msg)


class SimpleTokenizer(object):
    def __init__(self, stopwords=None, min_length=1, split_sym=None):
        if stopwords == 'default':
            stopwords = [
                'a', 'an', 'and', 'are', 'as', 'at', 'be', 'by', 'for',
                'from', 'has', 'he', 'in', 'is', 'its', 'of', 'on', 'or',
                'that', 'the', 'to', 'was ', 'were', 'will', 'with'
            ]
        elif stopwords is None:
            stopwords = []

        self.stopwords = set(stopwords)

        if split_sym is None:
            split_sym = []

        split_sym = punctuation + ''.join(split_sym)

        self.min_length = min_length
        self.re_tokenize = re.compile(r'&\w+;|\W+|_')

    def tokenize(self, text, lower=True):
        """Tokenize text"""
        if lower:
            text = text.lower()
        for tok in self.re_tokenize.split(text):
            if len(tok) >= self.min_length and tok not in self.stopwords:
                yield tok

    def tokenize_list(self, text, lower=True):
        if lower:
            text = text.lower()
        return [
            tok for tok in self.re_tokenize.split(text)
            if len(tok) >= self.min_length and tok not in self.stopwords
        ]


def db_key_encode(term):
    if six.PY2:
        return term
    else:
        return term.encode('utf-8')


def countlines(fn):
    """Count lines in fn. Slightly modified version of
    http://stackoverflow.com/a/27518377"""
    with open(fn, 'rb') as f:
        bufgen = takewhile(
            lambda x: x, (f.read(1024 * 1024) for _ in repeat(None)))
        ln = sum(buf.count(b'\n') for buf in bufgen)
    return ln


class SimstringDBWriter(object):
    def __init__(self, path):

        if not(os.path.exists(path)) or not(os.path.isdir(path)):
            err_msg = (
                '"{}" does not exists or it is not a directory.'
            ).format(path)
            raise IOError(err_msg)
        else:
            try:
                os.makedirs(path)
            except OSError:
                pass

        self.db = simstring.writer(
            prepare_string_for_db_input(
                os.path.join(path, 'umls-terms.simstring')
            ),
            3, False, True
        )

    def insert(self, term):
        term = prepare_string_for_db_input(safe_unicode(term))
        self.db.insert(term)


class SimstringDBReader(object):
    def __init__(self, path, similarity_name, threshold):
        if not(os.path.exists(path)) or not(os.path.isdir(path)):
            err_msg = (
                '"{}" does not exists or it is not a directory.'
            ).format(path)
            raise IOError(err_msg)

        self.db = simstring.reader(
            prepare_string_for_db_input(
                os.path.join(path, 'umls-terms.simstring')
            )
        )
        self.db.measure = getattr(simstring, similarity_name)
        self.db.threshold = threshold

    def get(self, term):
        term = prepare_string_for_db_input(safe_unicode(term))
        return self.db.retrieve(term)


class Intervals(object):
    def __init__(self):
        self.intervals = []

    def _is_overlapping_intervals(self, a, b):
        if b[0] < a[1] and b[1] > a[0]:
            return True
        elif a[0] < b[1] and a[1] > b[0]:
            return True
        else:
            return False

    def __contains__(self, interval):
        return any(
            self._is_overlapping_intervals(interval, other)
            for other in self.intervals
        )

    def append(self, interval):
        self.intervals.append(interval)


class CuiSemTypesDB(object):
    def __init__(self, path, database_backend='leveldb'):
        if not (os.path.exists(path) or os.path.isdir(path)):
            err_msg = (
                '"{}" is not a valid directory').format(path)
            raise IOError(err_msg)

        if database_backend == 'unqlite':
            assert UNQLITE_AVAILABLE, (
                'You selected unqlite as database backend, but it is not '
                'installed. Please install it via `pip install unqlite`'
            )
            self.cui_db = unqlite.UnQLite(os.path.join(path, 'cui.unqlite'))
            self.cui_db_put = self.cui_db.store
            self.cui_db_get = self.cui_db.fetch
            self.semtypes_db = unqlite.UnQLite(os.path.join(path, 'semtypes.unqlite'))
            self.semtypes_db_put = self.semtypes_db.store
            self.semtypes_db_get = self.semtypes_db.fetch
        elif database_backend == 'leveldb':
            self.cui_db = leveldb.LevelDB(os.path.join(path, 'cui.leveldb'))
            self.cui_db_put = self.cui_db.Put
            self.cui_db_get = self.cui_db.Get
            self.semtypes_db = leveldb.LevelDB(os.path.join(path, 'semtypes.leveldb'))
            self.semtypes_db_put = self.semtypes_db.Put
            self.semtypes_db_get = self.semtypes_db.Get
        else:
            raise ValueError(f'database_backend {database_backend} not recognized')

    def has_term(self, term):
        term = prepare_string_for_db_input(safe_unicode(term))
        try:
            self.cui_db_get(db_key_encode(term))
            return True
        except KeyError:
            return

    def insert(self, term, cui, semtypes, is_preferred):
        term = prepare_string_for_db_input(safe_unicode(term))
        cui = prepare_string_for_db_input(safe_unicode(cui))

        # some terms have multiple cuis associated with them,
        # so we store them all
        try:
            cuis = pickle.loads(self.cui_db_get(db_key_encode(term)))
        except KeyError:
            cuis = set()

        cuis.add((cui, is_preferred))
        self.cui_db_put(db_key_encode(term), pickle.dumps(cuis))

        try:
            self.semtypes_db_get(db_key_encode(cui))
        except KeyError:
            self.semtypes_db_put(
                db_key_encode(cui), pickle.dumps(set(semtypes))
            )

    def get(self, term):
        term = prepare_string_for_db_input(safe_unicode(term))
        try:
            cuis = pickle.loads(self.cui_db_get(db_key_encode(term)))
        except KeyError:
            cuis = set()

        matches = (
            (
                cui,
                pickle.loads(self.semtypes_db_get(db_key_encode(cui))),
                is_preferred
            )
            for cui, is_preferred in cuis
        )
        return matches



from .network import MinimalClient
from .core import QuickUMLS


def get_quickumls_client(host='localhost', port=4645):
    '''Return a client for a QuickUMLS server running on host at port'''
    client = MinimalClient(QuickUMLS, host=host, port=port, buffersize=4096)
    return client



from setuptools import setup, find_packages
import os
import io
import sys
import contextlib

PACKAGES = find_packages()

# From https://github.com/explosion/spaCy/blob/master/setup.py
@contextlib.contextmanager
def chdir(new_dir):
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        sys.path.insert(0, new_dir)
        yield
    finally:
        del sys.path[0]
        os.chdir(old_dir)


def setup_package():
    root = os.path.abspath(os.path.dirname(__file__))

    with open('README.md') as reader:
        readme = reader.read()

    with open('requirements.txt') as f:
        requirements = f.read().splitlines()

    dependency_links = []
    i = 0
    while i < len(requirements):
        if requirements[i].startswith('https://'):
            dependency_links.append(requirements.pop(i))
        else:
            i += 1


    # From https://github.com/explosion/spaCy/blob/master/setup.py
    with chdir(root):
        with io.open(os.path.join(root, "quickumls", "about.py"), encoding="utf8") as f:
            about = {}
            exec(f.read(), about)

    setup(
        name=about['__title__'],
        version=about['__version__'],
        description=(
            'QuickUMLS is a tool for fast, unsupervised biomedical '
            'concept extraction from medical text'
        ),
        packages=PACKAGES,
        long_description=readme,
        long_description_content_type='text/markdown',
        author=about['__author__'],
        author_email=about['__email__'],
        url='https://github.com/Georgetown-IR-Lab/QuickUMLS',
        license=about['__license__'],
        install_requires=requirements,
        dependency_links=dependency_links,
        classifiers=[
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 2",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Development Status :: 5 - Production/Stable",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
        ]
    )

if __name__ == '__main__':
    setup_package()



'''Minimal client server through sockets
https://github.com/lucasoldaini/MinimalServer'''

import six
import sys
import time
import math
import socket
import inspect
import datetime
import threading

try:
    import SocketServer as socketserver
    import cPickle as pickle
except ImportError:
    import socketserver
    import pickle


def pad_message(message, blocklength):
    """Pad a message so its length is a multiple of blocklength."""
    message_padded_length = (
        int(math.ceil(len(message) / blocklength)) * blocklength)
    padded_message = message.ljust(message_padded_length)
    return padded_message


def receive_data_in_chunks(sock, buffersize):
    """Receive data in chunks of size buffersize from the socket"""
    chunk = sock.recv(buffersize)
    chunks = [chunk]

    # keep reading until chunks are available
    while len(chunk.strip()):
        chunk = sock.recv(buffersize)
        chunks.append(chunk)

    data = b''.join(chunks).strip()
    return data


def send_data_in_chunks(data, sock, buffersize):
    """Split the message into chunks ans send it"""
    sock.sendall(pad_message(data, buffersize))

    # We sent an empty chunk to signal that we're done
    # transmitting the message.
    sock.send(b' ' * buffersize)


class MinimalServerHandler(socketserver.BaseRequestHandler):
    """Handle requests to the TCP server"""

    def handle(self):
        """Handle request by executing the specified method from the
        served object"""

        # receives data
        data = receive_data_in_chunks(self.request, self.server.buffersize)

        # parses the target method name, args, and kwargs
        method_name, args, kwargs = pickle.loads(data)

        # try executing the method from the server object, if it
        # fails, pass the error as response (the client will raise
        # the expection)
        try:
            response = getattr(
                self.server.served_object, method_name)(*args, **kwargs)
        except Exception as ex:
            response = ex

        # send the response to the client in chunks
        send_data_in_chunks(
            pickle.dumps(response, protocol=self.server.pickle_protocol),
            self.request, self.server.buffersize)


class MinimalServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """TCP Server"""
    served_object = None
    buffersize = 2048
    pickle_protocol = None


class MinimalClient(object):
    """Minimal client to provide communication with the server"""

    def __init__(self, target_class, host='localhost', port=4444,
                 buffersize=2048, pickle_protocol=None):
        """Initialize the client
        Args:
            target_class (object): the class to be served by the
                server. Only public methods will be binded.
            host (str or unicode): the host address of the server.
            port (int): the port the server is listening to.
            buffersize (int): the size of the buffer used for communication.
                Must be the same for both the server and the client.
            pickle_prtocol (int or None): the version number of the protocol
                used to pickle / unpickle objects. Necessary to be set
                if and only if server and client are running on different
                Python versions.
        """

        if pickle_protocol is None:
            pickle_protocol = pickle.HIGHEST_PROTOCOL

        self.host = host
        self.port = port
        self.buffersize = buffersize
        self.pickle_protocol = pickle_protocol

        if six.PY2:
            predicate = inspect.ismethod
        else:
            predicate = inspect.isfunction

        # bind public methods on target_class here
        for method_name, method in inspect.getmembers(
                target_class, predicate=predicate):
            if method_name.startswith('_'):
                continue
            setattr(self, method_name, self._func_req_wrapper(method_name))

    def _func_req_wrapper(self, method_name):
        """create a method with the same method_name that communicate
        with the server"""

        def func_request(*args, **kwargs):
            """Send the request to the server"""

            # prepare the data
            data = pickle.dumps((method_name, args, kwargs),
                                protocol=self.pickle_protocol)

            # open the socket
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))

            # tries sending the data and getting a response back
            try:
                send_data_in_chunks(data, sock, self.buffersize)
                response = receive_data_in_chunks(sock, self.buffersize)
            finally:
                sock.close()

            try:
                # unpickles the response
                data = pickle.loads(response)
            except EOFError:
                # server sent an empty message"
                msg = 'empty message received from the server.'
                raise RuntimeError(msg)

            # raises an exception if an exception was raised by the
            # served object while the server was executing method named
            # method_name
            if isinstance(data, Exception):
                raise data

            return data

        return func_request


def run_server(served_object, host='localhost',
               port=4444, buffersize=2048, pickle_protocol=None):
    """Runs the server
    Args:
        served_obkect (object): the object to be served by the
            server. Only public methods will be served.
        host (str or unicode): the host address of the server.
        port (int): the port the server is listening to.
        buffersize (int): the size of the buffer used for communication.
            Must be the same for both the server and the client.
        pickle_prtocol (int or None): the version number of the protocol
            used to pickle / unpickle objects. Necessary to be set
            if and only if server and client are running on different
            Python versions.
    """

    if pickle_protocol is None:
        pickle_protocol = pickle.HIGHEST_PROTOCOL

    # allows port reuse
    MinimalServer.allow_reuse_address = True

    # Initialize the server, set served_object, buffersize, pickle_prtocol
    server = MinimalServer((host, port), MinimalServerHandler)
    server.served_object = served_object
    server.buffersize = buffersize
    server.pickle_protocol = pickle_protocol

    # Start a thread with the server -- that thread will then start one
    # more thread for each request
    server_thread = threading.Thread(target=server.serve_forever)
    # Exit the server thread when the main thread terminates
    server_thread.daemon = True
    server_thread.start()
    print(
        '[{}] server running at {}:{} (press ^C to interrupt)'.format(
            datetime.datetime.now().isoformat(), host, port, server_thread.name
        )
    )

    # Wait for termination
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print(
            '\n[{}] server stopped'.format(datetime.datetime.now().isoformat())
        )

    # Terminate the server
    server.shutdown()
    server.server_close()
